*
 *
 * Copyright (c) IBM Corporation.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY IBM Corporation ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ARIZONA STATE UNIVERSITY
 * NOR ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * @author Dan He (dhe@us.ibm.com, danhe@cs.ucla.edu)
 * @version Revision: 1
 * @latest Date: 02/10/2014
 * 
 * Please cite the paper "Optimal Algorithms for Haplotype Assembly From Whole-Genome Sequence Data", 
 * Dan He, Arthur Choi, Knot Pipatsrisawat, Adnan Darwiche and Eleazar Eskin, Bioinformatics (2010) 26 (12): i183-i190.
 *
 *


import java.util.*;
import java.io.*;

public class OptimalHaplotype_Clean {
	
	public static final int LENGTHTHRESHOLD = 15; 
	public static final int HAPLOMAPLENGTH = 5000;
	public static final int LONGREADSNUM = 5000;

	
	
	public int pairedStartPos = Integer.MAX_VALUE;
	public String[] longReads = new String[LONGREADSNUM];
	public ArrayList<String> allReads = new ArrayList<String>();
	public ArrayList<String> justLongReads;
	public int longReadsNum = 0;
	public String fileName = "";
	public String optimalHap = "";
	
	public OptimalHaplotype_Clean(String fileName) {
		pairedStartPos = Integer.MAX_VALUE;
		justLongReads = new ArrayList<String>();
		this.fileName = fileName;
		optimalHap = "";
		allReads = new ArrayList<String>();
	}
	
	public static ArrayList<String> enumerateBinaryStrings(int length) {
		ArrayList<String> binaryStrings = new ArrayList<String>();
		
		if (length == 0) 
			return binaryStrings;
		else {
			LinkedList<String> queue = new LinkedList<String>();
			queue.add("0");
			queue.add("1");
			while (queue.size() > 0) {
				String top = queue.poll();
				if (top.length() == length) {
					binaryStrings.add(top);
				} else {
					queue.add(top+"0");
					queue.add(top+"1");
				}
			}
		}
		
		return binaryStrings;
	}
	

	
	/**
	 * dynamic programming algorithm considering only short reads
	 * @param shortReads: <starting position, array of reads starting from there>
	 */
	public int dynamicProgramming(TreeMap<Integer, ArrayList<String>> shortReads) {
		Iterator<Integer> iterator = shortReads.keySet().iterator();
		HashMap<String, String>[] pointer = new HashMap[HAPLOMAPLENGTH];
		//dynamic programming matrix
		HashMap<String, Integer>[] score = new HashMap[HAPLOMAPLENGTH];
		int matrixSize = 0;
		int matrixStart = 0;
		while (iterator.hasNext()) {
			int key = iterator.next();
			ArrayList<String> stringSet = shortReads.get(key);

			int maxLength = maxLength(stringSet);
			if (matrixSize == 0) {
				matrixStart = key;
				key = 1;
				ArrayList<String> binaryStrings = enumerateBinaryStrings(maxLength);
				score[key] = new HashMap<String, Integer>();
				pointer[key] = new HashMap<String, String>();
				for (int i = 0; i < binaryStrings.size(); ++i) {
					String binary = binaryStrings.get(i);
					
					int mecScore = mecScore(binary, stringSet);
					score[key].put(binary, mecScore);
					pointer[key].put(binary, "");
				}
			} else {
				key = key - matrixStart + 1;
				//compute the remaining matrix
				int index = key-1;
				ArrayList<String> binaryStrings;
				while (index >= 1 && score[index] == null)
					index--;
				if (index >= 1 && score[index] != null) {
					HashMap<String, Integer> temp = score[index];
					Iterator<String> iterator1 = temp.keySet().iterator();
					String firstInstance = iterator1.next();
					
					if (index + firstInstance.length() - 1 < key) {
						binaryStrings = enumerateBinaryStrings(maxLength);
						score[key] = new HashMap<String, Integer>();
						pointer[key] = new HashMap<String, String>();
						for (int i = 0; i < binaryStrings.size(); ++i) {
							String binary = binaryStrings.get(i);
							
							int mecScore = mecScore(binary, stringSet);
							int minMEC = Integer.MAX_VALUE;
							
							Iterator<String> tempIterator = score[index].keySet().iterator();
							String tempPointer = new String();
							while (tempIterator.hasNext()) {
								String tempKey = tempIterator.next();
								if (score[index].get(tempKey) + mecScore < minMEC) {
									minMEC = score[index].get(tempKey) + mecScore;
									tempPointer = index + "-" + tempKey;
								}
							}
							
							pointer[key].put(binary, tempPointer);
							score[key].put(binary, minMEC);
							
						}
					} else {
						int maxIndex = index + firstInstance.length() - 1 > key + maxLength - 1 ? 
								index + firstInstance.length() - 1 - key + 1 : maxLength;
						binaryStrings = enumerateBinaryStrings(maxIndex);
						score[key] = new HashMap<String, Integer>();
						pointer[key] = new HashMap<String, String>();
						for (int i = 0; i < binaryStrings.size(); ++i) {
							String binary = binaryStrings.get(i);
							
							int mecScore = mecScore(binary, stringSet);
							int minMEC = Integer.MAX_VALUE;
							//generate suffix of the string
							int suffixEndPos = index + firstInstance.length() - 1 > key + maxLength - 1 ?
									binary.length() : index + firstInstance.length() - 1 - key + 1;
							String suffix = binary.substring(0, suffixEndPos);
							
							// generate the prefix of the string
							int prefixLength = key - index;
							String tempPointer = new String();
							ArrayList<String> prefix = enumerateBinaryStrings(prefixLength);
							for (int j = 0; j < prefix.size(); ++j) {
								String finalString = prefix.get(j) + suffix;
								if (temp.containsKey(finalString) && temp.get(finalString) + mecScore < minMEC) {
									minMEC = temp.get(finalString) + mecScore;
									tempPointer = index + "-" + finalString;
								}
							}
							
							score[key].put(binary, minMEC);
							pointer[key].put(binary, tempPointer);
						}
					}
				}
			}
			matrixSize = key;
		}
		
		Iterator<String> iterator1 = score[matrixSize].keySet().iterator();
		int optimalScore = Integer.MAX_VALUE;
		String optimalEnd = new String(); 
		while (iterator1.hasNext()) {
			String key = iterator1.next();
			if (score[matrixSize].get(key) < optimalScore) {
				optimalScore = score[matrixSize].get(key);
				optimalEnd = key;
			}
		}
		
		
		System.out.println("score is: " + optimalScore);

		//trace back for the optimal haplotype
		int trackBack = matrixSize;
		String haplotype = optimalEnd;
		String current = optimalEnd;
		int currentPos = trackBack;
		while (trackBack >= 1 && pointer[trackBack] != null && pointer[trackBack].size() > 0) {
			String last = pointer[trackBack].get(current);
			currentPos = trackBack;
			if (last.equals("")) {
				break;
			} else {
				String[] split = last.split("-");
				trackBack = Integer.parseInt(split[0]);
				last = split[1];

				haplotype = combine(last, haplotype, trackBack + last.length() - 1, currentPos);

				current = last;
			}
		}
		System.out.println("haplotype is: " + haplotype + " of length :" + haplotype.length());
		this.optimalHap = haplotype;
		return optimalScore;
	}

	public int indicateDiff(String haplotype, String read, int hapStart, int readStart, int hapEnd, int readEnd, int lastEnd) {
		int indicateDiff = 0;
		int start = Math.max(hapStart, readStart);
		start = Math.max(start, lastEnd);
		int end = Math.min(hapEnd, readEnd);
		for (int i = start; i <= end; ++i) {
			char hapChar = haplotype.charAt(i-hapStart);
			char readChar = read.charAt(i-readStart);
			if ( hapChar != '-' && readChar != '-' && hapChar != readChar) {
				indicateDiff++;
			}
		}
		return indicateDiff;
	}
	
	public int indicateSame(String haplotype, String read, int hapStart, int readStart, int hapEnd, int readEnd, int lastEnd) {
		int indicateSame = 0; //same hap
 		int start = Math.max(hapStart, readStart);
		int end = Math.min(hapEnd, readEnd);
		start = Math.max(start, lastEnd);
		for (int i = start; i <= end; ++i) {
			char hapChar = haplotype.charAt(i-hapStart);
			char readChar = read.charAt(i-readStart);
			if ( hapChar != '-' && readChar != '-' && hapChar == readChar) {
				indicateSame++;
			}
		}
		return indicateSame;
	}
	
	/**
	 * the mec by assigning paired end reads to the optimal haplotype 
	 * computed by splitting the paired end reads
	 * @param shortReads
	 * @param haplotype: the optimal haplotype computed by splitting the paired end reads
	 * @return
	 */
	private int splitMec(TreeMap<Integer, ArrayList<String>> shortReads, String haplotype) {
		int mec = 0;
		Iterator<Integer> iterator = shortReads.keySet().iterator();
		int start = -1;
		while (iterator.hasNext()) {
			int key = iterator.next();
			if (start == -1) {
				start = key;
			} 
			key = key - start;
			System.out.println("key is: " + key);
			ArrayList<String> stringSet = shortReads.get(key);
			if (stringSet != null)
				mec += mecScore(haplotype.substring(key), stringSet);
			System.out.println("mec is: " + mec);
		}
		return mec;
	}
	
	/**
	 * combine two overlapping strings into one string
	 * @param last
	 * @param current
	 * @return
	 */
	private String combine(String last, String current, int lastEnd, int currentStart) {
		String combine = new String();
		if (lastEnd < currentStart) {
			combine = last;
			for (int i = lastEnd+1; i < currentStart; ++i) {
				combine += "-";
			}
			combine += current;
		} else {
			combine = last + current.substring(lastEnd+1-currentStart);
		}
		return combine;
	}
	
	/**
	 * compute the complent of a binary string
	 * @param binary
	 * @return
	 */
	public static String complement(String binary) {
		String complement = "";
		for (int i = 0; i < binary.length(); ++i) {
			if (binary.charAt(i) == '0') {
				complement += "1";
			} else if (binary.charAt(i) == '1'){
				complement += "0";
			} else {
				complement += "-";
			}
		}
		return complement;
	}
	
	/**
	 * Compute mismatch between a read and the haplotype
	 * @param haplotype
	 * @param s
	 * @return
	 */
	private int mismatchWithHaplotype(String haplotype, String s) {
		int mismatch = 0;
		int min = Math.min(haplotype.length(), s.length());
		for (int i = 0; i < min; ++i) {
			if (haplotype.charAt(i) != '-' && s.charAt(i) != '-' && haplotype.charAt(i) != s.charAt(i))
				mismatch++;
		}
		return mismatch;
	}
	
	public int mecScore(String haplotype, ArrayList<String> shortReads) {
		int score = 0;
		String complement = complement(haplotype);
		
		for (int i = 0; i < shortReads.size(); ++i) {
			String read = new String();
			String[] split = shortReads.get(i).split(" ");
			if (split.length == 2) {
				read = split[1];
			} else if (split.length == 4) { //short reads with gap or paired reads
				int readStart1 = Integer.parseInt(split[0]);
				int readStart2 = Integer.parseInt(split[2]);
				String read1 = split[1];
				String read2 = split[3];
				int pairedReadLength = readStart2 + read2.length() - 1 - readStart1 + 1;
				if (pairedReadLength <= haplotype.length()) { // short reads with gap
					int numGap = readStart2 - (readStart1 + read1.length() - 1) - 1;
					read = read1;
					for (int j = 0; j < numGap; ++j) {
						read += "-";
					}
					read += read2;
				} else { // paried end reads, only consider the first part
					read = read1;
				}
				
			}
			
			int m1 = mismatchWithHaplotype(haplotype, read);
			int m2 = mismatchWithHaplotype(complement, read);
			score += Math.min(m1, m2);
		}
		return score;
	}
	
	/**
	 * compute max length of the strings in a set
	 * @param stringSet
	 * @return
	 */
	public int maxLength(ArrayList<String> stringSet) {
		int maxLength = -1;
		for (int i = 0; i < stringSet.size(); ++i) {
			String binary = stringSet.get(i);
			String[] split= binary.split(" ");
			
			if (split.length == 2 && split[1].length() > maxLength && split[1].length() <= LENGTHTHRESHOLD) 
				maxLength = split[1].length();
			else if (split.length == 4) {
				int readStart1 = Integer.parseInt(split[0]);
				int readStart2 = Integer.parseInt(split[2]);
				String read1 = split[1];
				String read2 = split[3];
				int pairedReadLength = readStart2 + read2.length() - 1 - readStart1 + 1;
				if (pairedReadLength <= LENGTHTHRESHOLD && pairedReadLength > maxLength) 
					maxLength = pairedReadLength;
				else if (pairedReadLength > LENGTHTHRESHOLD && read1.length() <= LENGTHTHRESHOLD && read1.length() > maxLength) {
					maxLength = read1.length();
				}
			}
			
		}
		return maxLength;
	}
	
	/**
	 * read the file and return a tree map when paired end reads are considered
	 * @param fileName
	 * @return
	 */
	public TreeMap<Integer, ArrayList<String>> readFilePairedEnd(String fileName) {
		File matrixFile = new File(fileName);
		TreeMap<Integer, ArrayList<String>> reads = new TreeMap<Integer, ArrayList<String>>();

		try {
			BufferedReader input = new BufferedReader(new FileReader(matrixFile));
			String line = null; 
			int lineNum = 0;
			
			while ((line = input.readLine()) != null) {
				if (line.split("  ").length == 1) line = input.readLine();
				String[] split1 = line.split("  ");
				String[] split2 = split1[1].split(" ");
				if (split2.length == 2) { //regular reads
					int key = Integer.parseInt(split2[0]);
					
					if (split2[1].length() <= LENGTHTHRESHOLD) {
						if (reads.containsKey(key)) {
							reads.get(key).add(split1[1]);
						} else {
							ArrayList<String> stringSet = new ArrayList<String>();
							stringSet.add(split1[1]);
							reads.put(key, stringSet);
						}
					} else {
						longReads[longReadsNum] = split1[1];
						longReadsNum++;
					}
				} else if (split2.length == 4){ //paired end reads
					
					
					int readStart1 = Integer.parseInt(split2[0]);
					int readStart2 = Integer.parseInt(split2[2]);
					String read1 = split2[1];
					String read2 = split2[3];
			
					int pairedReadLength = readStart2 + read2.length() - 1 - readStart1 + 1;
					
					
					if (pairedReadLength <= LENGTHTHRESHOLD) { // short reads with gap
						if (reads.containsKey(readStart1)) {
							reads.get(readStart1).add(split1[1]);
						} else {
							ArrayList<String> stringSet = new ArrayList<String>();
							stringSet.add(split1[1]);
							reads.put(readStart1, stringSet);
						}
					} else {
						
						longReads[longReadsNum] = split1[1];
						longReadsNum++;
					}
				}
				lineNum++;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return reads;
	}
	
	/**
	 * slideWindow from start to end
	 * @param start
	 * @param end
	 */
	private void slideWindow(int start1, int end1) {
		//slide the window across the whole region and count the number of long reads in each window
		int max = 0;
		for (int i = start1; i < end1 - 15; ++i) {
			int start = i; 
			int end = i + 15 - 1;
			int num = 0;
			for (int j = 0; j < longReadsNum; ++j) {
				String read = longReads[j];
				String split[] = read.split(" ");
				if (split.length == 2) {
					if (Integer.parseInt(split[0]) >= start && (Integer.parseInt(split[0]) + split[1].length() - 1) <= end) {
						num++;
					}
				} else {
					int readStart1 = Integer.parseInt(split[0]);
					int readStart2 = Integer.parseInt(split[2]);
					String read1 = split[1];
					String read2 = split[3];
					if (readStart2 >= start && (readStart2 + read2.length() - 1) <= end) {
						num++;
					}
				}
			}
			if (num > max) {
				max = num;
			}
		}
		System.out.println("The size of window is : " + max);
	}
	
	/**
	 * read the file and return a tree map
	 * @param fileName
	 * @return
	 */
	public TreeMap<Integer, ArrayList<String>> readFile(String fileName) {
		File matrixFile = new File(fileName);
		TreeMap<Integer, ArrayList<String>> reads = new TreeMap<Integer, ArrayList<String>>();
		ArrayList<String> secondReads = new ArrayList<String>();
		int lineNumStart = 0;
		try {
			BufferedReader input = new BufferedReader(new FileReader(matrixFile));
			String line = null; 
			int lineNum = 0;
			int temp = 0;
			while ((line = input.readLine()) != null) {
				
				if (line.split("  ").length == 1) line = input.readLine();
				String[] split1 = line.split("  ");

				String[] split2 = split1[1].split(" ");
				if (split2.length == 2) { //regular reads
					int key = Integer.parseInt(split2[0]);
					this.allReads.add(lineNum, split1[1]);
					lineNum++;
					if (split2[1].length() <= LENGTHTHRESHOLD) {
						if (reads.containsKey(key)) {
							reads.get(key).add(split1[1]);
						} else {
							ArrayList<String> stringSet = new ArrayList<String>();
							stringSet.add(split1[1]);
							reads.put(key, stringSet);
						}
					} else {
						longReads[longReadsNum] = split1[1];
						longReadsNum++;
					}
				} else if (split2.length == 4) { //paired end reads
					
					
					int readStart1 = Integer.parseInt(split2[0]);
					int readStart2 = Integer.parseInt(split2[2]);
					String read1 = split2[1];
					String read2 = split2[3];
					String tempRead = read1 + " " + readStart2 + " " + read2;
					int length = readStart2 + read2.length() - 1 - readStart1 + 1;
					
					int pairedReadLength = readStart2 + read2.length() - 1 - readStart1 + 1;
					if (pairedReadLength <= LENGTHTHRESHOLD) { // short reads with gap
						if (reads.containsKey(readStart1)) {
							reads.get(readStart1).add(split1[1]);
						} else {
							ArrayList<String> stringSet = new ArrayList<String>();
							stringSet.add(split1[1]);
							reads.put(readStart1, stringSet);
						}				
						this.allReads.add(lineNum, split1[1]);
						lineNum++;
					} 

						longReads[longReadsNum] = split1[1];
						longReadsNum++;

					}
			}
			System.out.println("Total line is: " + lineNum);
			System.out.println("The long reads number is: " + longReadsNum);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return reads;
	}
	
	public static void main(String[] args) {
		String fileName = args[0];		
		
		OptimalHaplotype_Clean optimal = new OptimalHaplotype_Clean(fileName);

		TreeMap<Integer, ArrayList<String>> reads = optimal.readFilePairedEnd(optimal.fileName);
		optimal.dynamicProgramming(reads);
	}
}
